'use strict';

function removeListElement(state) {
  const input = state.input;
  const parent = input.parentNode;

  const list = parent.querySelector("plete-list");

  if (list) {
    parent.removeChild(list);
    delete state.list;
  }
}

function clearSuggestions(state) {
  state.selectedIndex = -1;
  removeListElement(state);
  state.input.removeAttribute("aria-controls");
  state.plete.setAttribute("aria-expanded", "false");
}

// https://github.com/angus-c/just/blob/master/packages/function-debounce/index.js
function debounce(fn, wait, callFirst) {
  var timeout;
  // eslint-disable-next-line consistent-return
  return function() {
    if (!wait) {
      // eslint-disable-next-line this/no-this
      return fn.apply(this, arguments);
    }
    // eslint-disable-next-line this/no-this
    var context = this;
    var args = arguments;
    var callNow = callFirst && !timeout;
    clearTimeout(timeout);
    // eslint-disable-next-line consistent-return
    timeout = setTimeout(function() {
      timeout = null;
      if (!callNow) {
        return fn.apply(context, args);
      }
    }, wait);

    if (callNow) {
      // eslint-disable-next-line this/no-this
      return fn.apply(this, arguments);
    }
  };
}

function highlight(state, index) {
  if (index >= 0) {
    const item = state.list.querySelectorAll("plete-item")[index];
    if (item) {
      item.classList.add("highlight");
    }
  }
}

function indexFrom(state, direction) {
  if (!state.list) {
    return -1;
  }

  const minIndex = 0;
  const maxIndex = state.list.children.length - 1;
  const newIndex = state.selectedIndex + direction;
  const actualIndex = Math.min(Math.max(newIndex, minIndex), maxIndex);

  return actualIndex;
}

function removeHighlight(state) {
  if (!state.list) {
    return;
  }

  const oldItem = state.list.querySelector("plete-item.highlight");
  if (oldItem) {
    oldItem.classList.remove("highlight");
  }
}

function moveHighlight(state, direction) {
  removeHighlight(state);

  const index = indexFrom(state, direction);

  highlight(state, index);
  state.selectedIndex = index;
}

/**
 * Normalizes the values of `event.key` across browsers, specifically aimed at
 * ensuring IE / Edge compatibility.
 *
 * The function only normalizes (and tests) the values that are actually used
 * in _this_ library. Other values also have inconsistencies across browsers,
 * they are ignored in this implementation in order to keep file size down.
 * *
 * @param  {String} key The value to normalize
 * @return {String}     A normalized value, the same across all browsers
 */
function normalizeKey(key) {
  if (key === "Up") {
    return "ArrowUp";
  }
  if (key === "Down") {
    return "ArrowDown";
  }
  if (key === "Esc") {
    return "Escape";
  }

  return key;
}

function select(state, index) {
  if (!state.list) {
    return;
  }

  if (index < 0) {
    state.select(undefined);
    return;
  }
  const selectedElement = state.list.children[index];
  const value = selectedElement.getAttribute("value");

  state.select(value);
  state.input.value = selectedElement.getAttribute("title");
  clearSuggestions(state);
}

const SPECIAL_KEYS = ["ArrowUp", "ArrowDown", "Enter", "Escape", "Tab"];

function handleInputKeydown(state, event) {
  const key = normalizeKey(event.key);

  if (!SPECIAL_KEYS.includes(key)) {
    return;
  }

  switch (key) {
    case "Tab":
      // if the list is not open or nothing is selected, just return here
      if (!state.list || state.selectedIndex === -1) {
        return;
      }
      select(state, state.selectedIndex);
      break;

    case "Enter":
      // if the list is not open, do nothing
      if (!state.list) {
        return;
      }

      // the suggestions are visible, but nothing is selected, prevent default
      if (state.selectedIndex === -1) {
        event.preventDefault();
        return;
      }

      select(state, state.selectedIndex);
      event.preventDefault();
      break;

    case "Escape":
      state.selectedIndex = -1;
      select(state, state.selectedIndex);
      clearSuggestions(state);
      break;

    case "ArrowUp":
    case "ArrowDown":
      moveHighlight(state, key === "ArrowUp" ? -1 : 1);
      event.preventDefault();
      break;

    /* istanbul ignore next This can't be tested as the function handles all special keys */
    default:
      throw new Error("Unhandled special key");
  }
}

function filter(state) {
  const inputValue = state.input.value;

  if (Array.isArray(state.dataSrc)) {
    // search for all the characters individually
    const string = inputValue
      .split("")
      .map(function(c) {
        return "\\" + c;
      })
      .join(".*");
    const regex = new RegExp(string, "i");

    if (typeof state.dataSrc[0] === "string") {
      return state.dataSrc.filter(v => regex.test(v));
    }

    return state.dataSrc.filter(v => {
      return regex.test(v.label);
    });
  }

  if (typeof state.dataSrc === "function") {
    return state.dataSrc(inputValue);
  }

  throw new Error("Unsupported dataSrc type");
}

function mouseDownHandler(event) {
  // Prevent the default mousedowm, which ensures the input is not blurred.
  // The actual selection will happen on click. This also ensures dragging the
  // cursor away from the list item will cancel the selection
  event.preventDefault();
}

/**
 * Used for generating DOM ids that will never repeat
 * @param  {string} prefix A prefix to use for the first part of the string
 * @return {string}        A DOM id
 */
function getNextId(prefix) {
  getNextId.id++;

  return `${prefix}-${getNextId.id}`;
}

getNextId.id = -1;

function addListElement(state, select) {
  const input = state.input;
  const parent = input.parentNode;

  removeListElement(state);

  const list = document.createElement("plete-list");
  list.setAttribute("id", getNextId("plete-list"));

  state.plete.setAttribute("aria-expanded", "true");

  state.input.setAttribute("aria-controls", list.getAttribute("id"));

  list.setAttribute("style", getStyles(input));
  list.setAttribute("aria-role", "listbox");

  if (state.cssClass) {
    list.setAttribute("class", state.cssClass);
  }
  parent.appendChild(list);

  list.addEventListener("mousedown", mouseDownHandler);

  // The click event is fired even if the corresponding mousedown event has called preventDefault
  list.addEventListener("click", function(event) {
    const element = event.target.closest("plete-item");

    // Only select on left click
    if (!element || event.button !== 0) {
      return;
    }

    event.preventDefault();

    const index = Array.from(
      element.closest("plete-list").querySelectorAll("plete-item")
    ).indexOf(element);

    select(state, index);
  });

  state.list = list;
}

function getStyles(input) {
  const left = input.offsetLeft;
  const top = input.offsetTop + input.offsetHeight;
  const width = input.offsetWidth;

  return `top: ${top}px; left: ${left}px; width: ${width}px;`;
}

function renderItem(state, value) {
  const item = document.createElement("plete-item");
  item.setAttribute("aria-role", "option");

  const valueIsString = typeof value === "string";
  const attrValue = valueIsString ? value : value.id;
  const label = valueIsString ? value : value.label;

  item.setAttribute("value", attrValue);
  item.setAttribute("title", label);

  if (typeof state.render === "function") {
    item.innerHTML = state.render(value);
  } else {
    item.textContent = label;
  }

  return item;
}

function render(state, data) {
  addListElement(state, select);
  const boundRenderItem = renderItem.bind(null, state);

  data.map(boundRenderItem).forEach(function(item) {
    state.list.appendChild(item);
  });

  state.selectedIndex = state.autoFirst ? 0 : -1;
  highlight(state, state.selectedIndex);
}

async function suggest(state) {
  clearSuggestions(state);
  state.selectedIndex = -1;
  select(state, state.selectedIndex);
  notifyBusy(state);

  const inputEmpty = state.input.value === "";
  const inputTooShort =
    typeof state.dataSrc === "function" &&
    state.input.value.length < state.minChars;

  if (inputEmpty || inputTooShort) {
    notifyReady(state);
    return;
  }

  abortRequest(state);

  const filterResult = filter(state);
  saveAbortFunction(state, filterResult);

  const filteredData = await getPromise(filterResult);

  const data = filteredData.slice(0, state.maxItems);

  render(state, data);
  notifyReady(state);
}

function getPromise(filterResult) {
  if (filterResult instanceof Promise) {
    return filterResult;
  }

  if (filterResult.promise instanceof Promise) {
    return filterResult.promise;
  }

  return filterResult;
}

function abortRequest(state) {
  if (!state.abortCurrentRequest) {
    return;
  }

  state.abortCurrentRequest();
  delete state.abortCurrentRequest;
}

function saveAbortFunction(state, filterResult) {
  if (!filterResult.abort) {
    return;
  }

  state.abortCurrentRequest = filterResult.abort;
}

function notifyBusy(state) {
  if (typeof state.busy === "function") {
    state.busy();
  }
}

function notifyReady(state) {
  if (typeof state.ready === "function") {
    state.ready();
  }
}

function attachEventListeners(state) {
  const debounceDelay = typeof state.dataSrc === "function" ? 120 : 0;
  const debouncedSuggest = debounce(suggest.bind(null, state), debounceDelay);

  state.input.addEventListener("keydown", handleInputKeydown.bind(null, state));
  state.input.addEventListener("input", debouncedSuggest);
  state.input.addEventListener("paste", suggest.bind(null, state));
  state.input.addEventListener("blur", clearSuggestions.bind(null, state));
}

function validateOptions(state) {
  const invalidInput =
    !(state.input instanceof HTMLElement) ||
    state.input.tagName.toLowerCase() !== "input" ||
    state.input.type !== "text";
  if (invalidInput) {
    throw new TypeError(
      `input option expected to be an input DOM element, but was: ${state.input}, with type ${state.input.type}`
    );
  }

  const invalidDataSrc = !(
    Array.isArray(state.dataSrc) || typeof state.dataSrc === "function"
  );
  if (invalidDataSrc) {
    throw new TypeError("dataSrc option expect to be an Array or a Function");
  }

  if (typeof state.autoFirst !== "boolean") {
    throw new TypeError("Expected autoFirst option to be a boolean");
  }

  if (typeof state.minChars !== "number" || state.minChars < 1) {
    throw new TypeError("Expected minChars option to be a positive integer");
  }

  if (typeof state.maxItems !== "number" || state.maxItems < 1) {
    throw new TypeError("Expected maxItems option to be a positive integer");
  }

  if (!["undefined", "function"].includes(typeof state.render)) {
    throw new TypeError("Expected render option to be a function or undefined");
  }

  if (!["undefined", "function"].includes(typeof state.busy)) {
    throw new TypeError("Expected busy option to be a function or undefined");
  }

  if (!["undefined", "function"].includes(typeof state.ready)) {
    throw new TypeError("Expected ready option to be a function or undefined");
  }

  if (!["undefined", "function"].includes(typeof state.select)) {
    throw new TypeError("Expected select option to be a function or undefined");
  }

  if (!["undefined", "string"].includes(typeof state.cssClass)) {
    throw new TypeError("Expected cssClass option to be a string or undefined");
  }
}

function ensurePleteElement(state) {
  state.plete = state.input.closest("plete");

  if (!state.plete) {
    state.plete = document.createElement("plete");
    state.input.parentNode.appendChild(state.plete);
    state.plete.appendChild(state.input);
  }

  state.plete.setAttribute("aria-role", "combobox");
  state.plete.setAttribute("aria-expanded", "false");
}

function setAutoComplete(state) {
  state.input.setAttribute("aria-autocomplete", "list");
}

const DEFAULT_OPTIONS = {
  autoFirst: true,
  maxItems: 5,
  minChars: 3,
  render: undefined
};

function Plete(options) {
  // eslint-disable-next-line this/no-this
  if (!(this instanceof Plete)) {
    throw new Error("Plete should be called with the `new` keyword");
  }

  const state = Object.assign({}, DEFAULT_OPTIONS, options);

  validateOptions(state);
  setAutoComplete(state);
  ensurePleteElement(state);
  attachEventListeners(state);
}

module.exports = Plete;
